<?php

/**
 * @file
 * Admin functionality for Drunkins.
 */

function drunkins_overview() {
  module_load_include('inc', 'drunkins', 'drunkins.run');

  // Note: (batch) job id == queue name.

  // TODO: check where t() is needed

  $header = array('Queue / Job', 'Type', 'Status', '# items left', 'Last start', 'Last finish', 'Operations');
  $rows = array();

  // Compile info from two sources which do not fully overlap:
  // - jobs info (including status/run info)
  // - all queues from hook_queue_info (not only the Drunkins job queues).
  // (In D7 there's probably no difference in 'all queues'. because hardly
  // anything implements hook_queue_info. Still, we will accommodate for it.)
  $jobs_info = drunkins_job_status_info();
  $queues = module_invoke_all('queue_info');
  drupal_alter('queue_info', $queues);
  foreach ($queues as $queue_name => $queue_info) {
    $runner_settings = _drunkins_get_process_config($queue_name);
    $queue = DrupalQueue::get($queue_name);
    $count = $queue->numberOfItems();
    $countinfo = $count;
    $title = empty($queue_info['title']) ? $queue_name : ($queue_info['title'] . ' (<em>' . $queue_name . '</em>)');

    if (isset($queue_info['drunkins'])) {
      $options = array();
      if (!isset($queue_info['drunkins']['cron_schedule'])) {
        $options[] = 'cron disabled';
      }
      $type = 'Drunkins'
              . ($options ? ' (' . implode('; ', $options) . ')' : '');
      if ($queue_info['drunkins']['running'] < 2) {
        $status = empty($queue_info['drunkins']['started']) ? '-' : 'finished';
        if ($queue_info['drunkins']['running'] == 1) {
          $status .= '<br/><em>(cron due'
                     . (!empty($queue_info['drunkins']['run_next'])
              ? ' since ' . format_date($queue_info['drunkins']['run_next'], 'short') : '')
                     . ')</em>';
        }
        // finished => there should be no queued items
        if ($count) {
          $countinfo .= ' <strong>(?!)</strong>';
        }
      }
      else {
        $status_detail = array($queue_info['drunkins']['method']);

        $context = isset($jobs_info[$queue_name]['process_context']) ? $jobs_info[$queue_name]['process_context'] : array();
        if (!empty($context['drunkins_process_more'])) {
          $start_runs = empty($context['drunkins_rerun_start']) ? '?' : $context['drunkins_rerun_start'];
          $start_count = isset($context['drunkins_rerun_start_count']) ? $context['drunkins_rerun_start_count'] : '?';
          $status_detail[] = "fetching more after {$start_runs} runs/{$start_count} items";
        }

        if ($queue_info['drunkins']['running'] == 3) {
          if ($queue_info['drunkins']['method'] === 'batch') {
            // There should be no queued items for a 'batch' queue job.
            // (They should be queued under another name.)
            if ($count) {
              $countinfo .= ' <strong>(?!)</strong>';
            }
            else {
              $countinfo = '?';
            }
          }
          else {
            if (!empty($runner_settings['force_cron'])) {
              $status_detail[] = 'cron forced for the current process';
            }
            if (!$count && empty($context['drunkins_process_more'])) {
              $status_detail[] = 'waiting to finish';
            }
          }
        }
        else {
          if ($queue_info['drunkins']['run_info'] == 4) {
            // It's debatable whether we should call this "running" in the UI?
            $status_detail[] = 'error on finish';
          }
          else {
            $status_detail[] = 'locked; cause unknown';
          }
        }

        $status = 'running (' . implode(', ', $status_detail) . ')';

        if (!empty($queue_info['drunkins']['cron_schedule'])) {
          // We have not calculated run_next yet, but should display it.
          _drunkins_calculate_dynamic_info($queue_info['drunkins'], $queue_name, TRUE);
        }
      }
      // Display next cron run, unless we already have.
      if ($queue_info['drunkins']['running'] != 1
          && !empty($queue_info['drunkins']['run_next'])) {
        $next = format_date($queue_info['drunkins']['run_next'], 'short');

        $status .= $queue_info['drunkins']['run_next'] >= REQUEST_TIME
          ? "<br/><em>(cron due $next)</em>" : "<br/><em>(cron due since $next)</em>";
      }

      $started = empty($queue_info['drunkins']['started']) ? '-' : format_date($queue_info['drunkins']['started']);
      $finished = empty($queue_info['drunkins']['finished']) ? '-' : format_date($queue_info['drunkins']['finished']);
      if ($finished != '-') {
        if ($queue_info['drunkins']['started'] <= $queue_info['drunkins']['finished']) {
          // Output info about previous process.
          $finished .= ' (' . $queue_info['drunkins']['method'];
          switch ($queue_info['drunkins']['run_info']) {
            case 3:
              // This very likely won't happen; {finished} is not updated on
              // "error_on_finish" in _drunkins_finish_process(). We're just
              // including it to be future proof and to not trigger the default.
              $finished .= '; error on finish)';
              break;
            case 4:
              $finished .= '; terminated on start)';
              break;
            case 5:
              $finished .= '; reset manually)';
              break;
            case 0:
              $finished .= ')';
              break;
            default:
              // Other states always mean 'running', so something's wrong:
              // finished should be smaller than started.
            $finished .= ' (<em>data error</em>)';
          }
        }
        else {
          // 'Last finish' is not connected with 'last start'. Do we need to specify this?
          $finished = '<em>(' . $finished . ')</em>';
        }
      }
      $actions = array(l('Start job / view data', "admin/config/workflow/drunkins/$queue_name"));
    }
    else {
      // Non-Drunkins queue
      // Check what we know from looking at queue_ui code
      $type = (isset($queue_info['batch']) || isset($queue_info['cron']['callback']))
        ? 'queue_ui' : '??';
      $status = $started = $finished = '?';
      // We can move queue_ui actions here, if we want... see also commented code above.
      $actions = array();
    }

    $rows[] = array($title, $type, $status, $countinfo, $started, $finished,
      implode(' - ', $actions)
    );
  }

  // Get queue table info for later.
  $other_queues = db_query("SELECT name, count(created) AS cnt, min(created) as crt FROM {queue} GROUP BY name")->fetchAllAssoc('name', PDO::FETCH_ASSOC);

  // D7 core still has separate cron queues. Our Drunkins jobs also have them,
  // so don't process these again. Since we're not sure if a queue uses the
  // {queue} table, we get the count through the official method. For 'started'
  // there is no official method, so we use the above if we have it.
  $cron_queues = module_invoke_all('cron_queue_info');
  drupal_alter('queue_info', $cron_queues);
  $cron_queues = array_diff_key($cron_queues, $queues);
  foreach ($cron_queues as $queue_name => $queue_info) {
    $title = empty($queue_info['title']) ? $queue_name : ($queue_info['title'] . ' (<em>' . $queue_name . '</em>)');
    $count = DrupalQueue::get($queue_name)->numberOfItems();
    $rows[] = array($title, 'Cron queue', $count ? 'running' : 'finished', $count,
      isset($other_queues[$queue_name]['crt']) ? format_date($other_queues[$queue_name]['crt']) : '?',
      '?', '');
  }

  // For fun, see what other queues we have in the {queue} table.
  // Instead of executing DrupalQueue::get($queue_name)->numberOfItems() for
  // each queue, we might as well get the count from the table immediately.
  $other_queues = array_diff_key($other_queues, $queues, $cron_queues);
  foreach ($other_queues as $queue_name => $queue_info) {
    $rows[] = array($queue_name, 'queue table', $queue_info['cnt'] ? 'running' : 'finished',
      $queue_info['cnt'], format_date($queue_info['crt']), '?', '');
    // TODO We can probably reset/remove leases here...
  }

  return array(
    '#theme' => 'table',
    '#header' => $header,
    '#rows' => $rows,
  );
}

/**
 * Title callback for various administrative screens.
 * @see drunkins_menu()
 */
function drunkins_title_callback($title, $job_id) {
  $job_info = drunkins_job_status_info($job_id);
  return str_replace(
    '%job_id',
    "'" . (isset($job_info['title']) ? $job_info['title'] : $job_id) . "'",
    $title);
}

/**
 * Form definition: display process (meta)data for a job.
 *
 * This is not really a form, there is no submit functionality - we just make it
 * one so we can use a textarea.
 */
function drunkins_view_job_metadata($form, &$form_state, $job_id) {
  if (empty($job_id)) {
    drupal_set_message(t('Cannot view process metadata for empty job!'), 'error');
  }
  $info = drunkins_job_status_info($job_id);
  if (empty($info)) {
    drupal_set_message(t('Unknown job name %process!', array('%process' => $job_id)), 'error');
  }

  // Process $info. process_settings contain everything in settings so we can
  // delete the latter, but we'll use array_diff() to be future compatible.
  if (isset($info['process_settings'])) {
    $info['settings'] = array_diff_key($info['settings'], $info['process_settings']);
    if (empty($info['settings'])) {
      unset($info['settings']);
    }
  }
  if (isset($info['process_runner_settings'])) {
    $info['runner_settings'] = array_diff_key($info['runner_settings'], $info['process_runner_settings']);
    if (empty($info['runner_settings'])) {
      unset($info['runner_settings']);
    }
  }

  // print_r is better on the eye than var_dump. However, it prints 1/''
  // for TRUE/FALSE, which can cause confusion.
  array_walk_recursive($info, function (&$value) {
    if (is_bool($value)) {
      $value = $value ? 'True' : 'False';
    }
  });
  $form['metadata'] = array(
    '#type'          => 'textarea',
    '#rows'          => 20,
    '#weight' => -99,
    '#default_value' => print_r($info, TRUE),
  );

  return $form;
}

/**
 * Form definition: display queued items for a job.
 *
 * This is not really a form, there is no submit functionality - we just make it
 * one so we can use a textarea.
 */
function drunkins_view_job_items($form, &$form_state, $job_id) {
  $items = NULL;
  if (empty($job_id)) {
    drupal_set_message(t('Cannot view process metadata for empty job.'), 'error');
  }
  $info = drunkins_job_status_info($job_id);
  if (empty($info)) {
    drupal_set_message(t('Unknown job name %process.', array('%process' => $job_id)), 'error');
  }
  else {
    /** @var DrupalReliableQueueInterface $queue */
    $queue = DrupalQueue::get($job_id);
    // $queue has no methods to peek at items...
    if (!$queue instanceof SystemQueue) {
      drupal_set_message(t('Cannot view %process items which are in a %queue_name queue.', array('%process' => $job_id, '%queue_name' => get_class($queue))), 'error');
    }
    else {
      // Max 100
      $items = db_query_range('SELECT item_id, created, expire, data FROM {queue} q WHERE name = :name ORDER BY item_id ASC', 0, 100, array(':name' => $job_id))->fetchAllAssoc('item_id');
      if (count($items) == 100) {
        drupal_set_message(t('The number of items is shown is (possibly) truncated to @number.', array('@number' => 100)), 'warning');
      }
      foreach ($items as &$item) {
        unset($item->item_id);
        $item->data = unserialize($item->data);
      }
      // print_r is better on the eye than var_dump. However, it prints 1/''
      // for TRUE/FALSE, which can cause confusion.
      array_walk_recursive($items, function (&$value) {
        if (is_bool($value)) {
          $value = $value ? 'True' : 'False';
        }
      });
    }
  }
  $form['metadata'] = array(
    '#type'          => 'textarea',
    '#rows'          => 20,
    '#weight' => -99,
    '#default_value' => print_r($items, TRUE),
  );

  return $form;
}

/**
 * Displays (or saves in Excel format) a list of items retrieved by the job.
 *
 * The 'output_format' setting is used by this function; a value of 'xls' will
 * output items in Excel format (and set HTTP headers for download.)
 * 'list_format' is commonly used by jobs' start() method as a hint for altering/
 * reformatting items for listing rather than processing; will be set to 'raw'
 * by default.
 * It is up to the individual job to (not) implement this if desired. (In other
 * words, this function does not guarantee that it will output useful data.)
 *
 * @param string $job_id
 *   ID for the job (which is equal to the queue name for queue processes),
 * @param array $settings
 *   Settings which should be passed to the job, which start() can then use for
 *   any purpose.
 *
 * @return string
 *   The output.
 *
 * @todo run start() and queue items _before_ redirecting?
 *    (Haven't thought that one through totally yet, but would solve
 *   out-of-memory issues in some cases and possibly prevent possible DDoS?
 *   Would also enable running-after-showing, if you keep the items.)
 */
function drunkins_list_items($job_id, array $settings = array()) {

  // If settings were not explicitly passed, construct them from URL arguments.
  if (empty($settings)) {
    $settings = array_diff_key($_GET, array('q' => 1));
    if (!isset($settings['list_format'])) {
      $settings['list_format'] = 'raw';
    }
  }
  elseif (!array($settings)) {
    $settings = array('list_format' => 'raw');
  }
  // Only do drupal_set_message() for errors.
  // We'd officially need to set logging in the 'runner settings' and
  // drunkins_get_job() would then move it to settings for us. But let's
  // 'cheat' and be more direct in this case by setting this unofficial
  // logger setting, since we don't need a runner. Maybe look at this later.
  $settings['logging'] = DrunkinsJob::LOG_SCREEN;

  module_load_include('inc', 'drunkins', 'drunkins.run');
  $job = drunkins_get_job($job_id, $settings);
  if ($job) {

    $context = array();
    $items = array();
    do {
      try {
        unset($context['drunkins_process_more']);
        $items = array_merge($items, $job->start($context));
      }
      catch (Exception $e) {
        // This could be any kind of exception. Its string representation could
        // give us more relevant info than just the message. However it often also
        // contains a whole backtrace (whose usefulness is limited) which will be
        // ugly if we just print it. On top of that, an exception message could be
        // unsafe, so we have to escape it, so we can't just replace newlines with
        // <br>. The next best thing is escaping the value and enclosing it in
        // <pre>.
        drupal_set_message('<pre>' . check_plain((string) $e) . '</pre>', 'error');
        drupal_goto('admin/config/workflow/drunkins/' . $job_id);
      }
      if (!is_array($items)) {
        drupal_set_message(t('Process could not start: @error', array('@error' => $items)), 'error');
        drupal_goto('admin/config/workflow/drunkins/' . $job_id);
      }

      if (!empty($context['drunkins_process_more'])) {
        // 'display_rerun' is probably a temporary solution, to be phased out
        // when we start working with queued items, as per above @todo.
        if (!empty($settings['display_rerun'])) {
          // Update context for the next call to start(). (We unfortunately
          // can't call _drunkins_run_start() because that changes run-state.)
          if (!empty($context['drunkins_rerun_start']) && is_int($context['drunkins_rerun_start'])) {
              $context['drunkins_rerun_start']++;
          } else {
              $context['drunkins_rerun_start'] = 1;
          }
          $context['drunkins_rerun_start_count'] = count($items);
          // Omit a message so that we can see where we time out, if we do.
          drupal_set_message(t('Fetched total @total items in @rerun_count iterations; trying to fetch more...',
            array('@rerun_count' => $context['drunkins_rerun_start'], '@total' => $context['drunkins_rerun_start_count'])));
        }
        else {
          drupal_set_message(t('More items need to be fetched, which is not supported here. Showing only the first batch.'), 'warning');
          if (isset($settings['cache_items']) && $settings['cache_items'] == 1) {
            drupal_set_message(t('Only a partial data set was cached. <strong>Please be careful about using the currently cached items for running a job.</strong> You may want to delete them to prevent confusion.'), 'warning');
          }
          // Break out of the loop.
          $context['drunkins_process_more'] = FALSE;
        }
      }
    } while (!empty($context['drunkins_process_more']));

    if (empty($items)) {
      drupal_set_message(t('There are no items to process.'));
      drupal_goto('admin/config/workflow/drunkins/' . $job_id);
    }
    else {
      $header = array();

      // Construct the table header. We need to loop through all items because
      // not all data may contain all fields.
      $prev_item = '';
      foreach ($items as $item) {
        if (is_array($item)) {
          // Item is array.
          $prev_key = '';
          foreach ($item as $key => $value) {
            if (!in_array($key, $header)) {
              if (!$prev_item || !$prev_key) {
                $header[] = $key;
              }
              else {
                // We're inserting a header for a field which was not present in
                // the previous item. Insert this header just after the previous
                // field. (This is not perfect because if two elements are
                // inserted behind the same $prev_key in different rows, they
                // may end up reordered. But it's better / the table may look
                // more logical, than if we just insert it at the end.)
                array_splice($header, array_search($prev_key, $header), 1, array($prev_key, $key));
              }
            }
            $prev_key = $key;
          }
        }
        else {
          // Item is scalar. Probably all of them are, but loop through items
          // to be sure.
          if (!in_array('Single value', $header)) {
            $header[] = 'Single value';
          }
        }
        $prev_item = $item;
      }

      if (!empty($settings['output_format'])
          && $settings['output_format'] == 'xls') {

        header("Content-type: application/vnd.ms-excel");
        header("Content-Disposition: attachment; filename=items.xls");
        header('Content-Transfer-Encoding: binary');
        header("Pragma: no-cache");
        header("Expires: 0");

        //chr(255) . chr(254) . mb_convert_encoding($table, 'UTF-16LE', 'UTF-8');
        print '<table><thead><tr><th>' . implode('</th><th>', $header) . '</th></tr></thead><tbody>';
        foreach ($items as $item) {
          $row = _drunkins_construct_list_row($item, $header, $settings);
          print '<tr><td>' . implode('</td><td>', $row) . '</td></tr>';
        }
        print '</tbody></table>';

      }
      else {
        $rows = array();
        $rows['header'] = $header;

        // Collect items.

        // Items are not check_plain()ed because they may be reformatted to
        // contain HTML. (Safe content is the job's responsibility.)

        foreach ($items as $item) {
          $rows['rows'][] = _drunkins_construct_list_row($item, $header, $settings);
        }

        $table = theme('table', $rows);

        return t('Displaying @count items.', array('@count' => count($rows['rows'])))
               . ' - '
               . l('.xls download', 'admin/config/workflow/drunkins/' . $job_id . '/display',
          array('query' => array_merge($settings, array('output_format' => 'xls'))))
               . ' - '
               . l('Job settings', 'admin/config/workflow/drunkins/' . $job_id,
            array('query' => $settings))
               . $table;
      }
    }
  }
  else {
    drupal_set_message(t('Invalid job id %job', array('%job' => $job_id)), 'error');
  }
}

function _drunkins_construct_list_row($item, array $header, array $settings) {
  $row = array();
  if (is_array($item)) {
    foreach ($header as $key) {
      if (!isset($item[$key])) {
        $row[] = '-';
      }
      elseif (is_array($item[$key])) {
        // We may do this entirely differently later. This is a quick hack.
        if (!empty($settings['list_output_arrays'])) {
          $row[] = check_plain(json_encode($item[$key]));
        }
        else {
          $row[] = '&lt;array&gt;';
        }
      }
      elseif (is_object($item[$key])) {
        // We may do this entirely differently later. This is a quick hack.
        if (!empty($settings['list_output_objects'])) {
          $row[] = check_plain(json_encode($item[$key]));
        }
        else {
          $row[] = '&lt;object&gt;';
        }
      }
      elseif (is_bool($item[$key])) {
        $row[] = '<em>' . ($item[$key] ? 'true' : 'false') . '</em>';
      }
      else {
        // @todo make it the responsibility of a UI class to override this if it wants to print safe values.
        $row[] = check_plain($item[$key]);
      }
    }
  }
  else {
    $row['Single value'] = check_plain($item);
  }

  return $row;
}

/// Batch functionality.

/**
 * Form definition for starting the batch update process.
 */
function drunkins_batch_form($form, $form_state, $job_id) {

  // We'd officially need to set logging in the 'runner settings' and
  // drunkins_get_job() would then move it to settings for us. But let's 'cheat'
  // and be more direct in this case by setting this unofficial logger setting,
  // since we don't need a runner. Maybe look at this later. (Note we will have
  // to keep passing an argument, even if empty array, to prevent getting the
  // settings for a running process returned.)
  module_load_include('inc', 'drunkins', 'drunkins.run');
  $job = drunkins_get_job($job_id, array('logging' => DrunkinsJob::LOG_SCREEN));
  if (!$job) {
    drupal_set_message(t('Invalid job id %job', array('%job' => $job_id)), 'error');
    return drunkins_overview();
  }
  $job_info = drunkins_job_status_info($job_id);

  $form = method_exists($job, 'settingsForm') ? $job->settingsForm() : array();

  // We allow influencing default values for known selectboxes by URL args.
  foreach (array('list_format', 'output_format', 'cache_items') as $element) {
    if (isset($_GET[$element])) {
      _drunkins_set_option_default($form, $element);
    }
  }

  $form['drunkins_job_id'] = array(
    '#type'  => 'value',
    '#value' => $job_id,
  );
  $form['cron_group'] = array(
    '#type'   => 'fieldset',
    '#collapsible' => TRUE,
    '#collapsed' => TRUE,
    '#title' => t('Delayed processing'),
    '#weight' => 100,
    '#access' => $job_info['running'] < 2,
  );
  if (!isset($job_info['cron_schedule'])) {
    $form['cron_group']['force_cron'] = array(
      '#type'   => 'checkbox',
      '#title'  => t('Force cron processing of items'),
      '#description' => t('This will have all enqueued items be processed on cron, this one time (despite cron being disabled for this job normally). If you don\'t select this option, a system outside of this module will need to process the queued items.'),
      '#default_value' => TRUE,
    );
  }
  $form['cron_group']['cron'] = array(
    '#type'   => 'submit',
    '#value'  => t('Enqueue'),
    '#suffix' => '<div class="description">This will place all items in a queue to be processed later.</div>',
  );
  $form['submit'] = array(
    '#type'   => 'submit',
    '#value'  => t('Start batch'),
    '#suffix' => '<div class="description">This will process the items immediately using Drupal\'s batch mode.</div>',
    '#weight' => 99,
    '#access' => $job_info['running'] < 2,
  );

  if ($job_info['running'] > 1) {
    $queue = DrupalQueue::get($job_id);
    $nr_items = $queue->numberOfItems();
    if (empty($form_state['input'])) {
      drupal_set_message(t('Job %job is !status in @method method, with !n items still to be processed; it cannot be started now.',
        array(
          '%job' => $job_id,
          '!status' => $job_info['running'] == 3 ? 'running' : 'locked',
          '@method' => $job_info['method'],
          '!n' => $nr_items
        )), 'warning');
    }
    if (user_access('access drunkins advanced')) {
      if ($nr_items) {
        // There is no generalized method for getting the number of _claimed_
        // items. Now that we're querying directly, we can also get min(expire)
        // for some extra safety. Note this is very much tied to SystemQueue, and
        // a rewrite will need to take that into account.
        $row = db_query('SELECT COUNT(expire) as amt, MAX(expire) as max_exp FROM {queue} WHERE name = :name and expire <> 0', array(':name' => $job_id))->fetchAssoc();
        if (empty($form_state['input'])) {
          if ($row['amt']) {
            drupal_set_message(t('@n of the queued items are currently claimed; their expiry date is @date or earlier.', array(
              '@n' => $row['amt'],
              '@date' => date('r', $row['max_exp'])
            )), 'warning');
          }
        }
      }
      if (empty($form_state['input'])) {
        drupal_set_message(t('See "Advanced actions" below.'), 'warning');
      }
      $form['drunkins_advanced'] = array(
        '#type'   => 'fieldset',
        '#collapsible' => TRUE,
        '#collapsed' => TRUE,
        '#title' => t('Advanced actions'),
        '#weight' => 101,
      );
      $parameters = array();
      if ($job_info['method'] === 'batch') {
        $addendum = ". A batch process' items cannot be deleted through here. The job will be 'unlocked' but if the batch process is currently still being run by another client, this has unknown consequences.";
      }
      else {
        $addendum = ' and delete @amt items.';
        $parameters['@amt'] = $nr_items;
      }
      $form['drunkins_advanced']['reset_job'] = array(
        '#type'   => 'submit',
        '#value'  => t('Reset job'),
        '#suffix' => '<div class="description">' . t("<strong>This will destroy your running job</strong>$addendum", $parameters) . '</div>',
      );
      if ($nr_items && $row['amt']) {
        $form['drunkins_advanced']['unclaim_items'] = array(
          '#type'   => 'submit',
          '#value'  => t('Un-claim items'),
          '#suffix' => '<div class="description">' . t("This will free up the @n claimed items up to be processed again. Only do this if you are sure they are not being processed anymore right now.", array('@n' => $row['amt'],)) . '</div>',
        );
        $form['drunkins_advanced']['unclaim_items_max'] = array(
          '#type' => 'value',
          '#value' => $row['max_exp'],
        );
      }
    }
  }
  elseif (!empty($job_info['runner_settings']['enqueue_by_fetcher'])) {
    // Temporary (hopefully) message. @todo fix this.
    drupal_set_message('<strong>WARNING:</strong> This job (which enqueues items externally to the module) currently <strong>does not support the "Start batch" and "Display data" command</strong>. The module should be extended for this.', 'warning');
    unset($form['submit']);
  }
  if (user_access('access drunkins advanced')) {
    // Inconsistent UI: below is a link; "Display" is a button. (Also, that's
    // in the job class instead of here)
    // @todo unify sometime?
    $form['display_process_data'] = array(
      '#markup' => l(t('Display metadata used by the current process.'), "admin/config/workflow/drunkins/$job_id/metadata"),
      '#weight' => 200,
      '#prefix' => '<div class="description">',
      '#suffix' => '</div>',
    );
    // Maybe fix up later, show in UI overview too?
    $form['display_process_items'] = array(
      '#markup' => l(t('Display items queued by the current process.'), "admin/config/workflow/drunkins/$job_id/items"),
      '#weight' => 201,
      '#prefix' => '<div class="description">',
      '#suffix' => '</div>',
    );
  }

  return $form;
}

/**
 * Set a default value of an 'options' form element, which is somewhere in
 * $form (possibly inside a fieldset)
 *
 * @param array $form
 *  Form definition
 * @param $element_name
 *  Element name. $_GET[$element_name] must hold the value to set.
 */
function _drunkins_set_option_default(&$form, $element_name) {
  if (isset($form[$element_name])) {

    // Found. Check whether this is a valid default value; if so, set.
    if (isset($form[$element_name]['#options'][$_GET[$element_name]])) {
      $form[$element_name]['#default_value'] = $_GET[$element_name];
    }
  }
  else {

    // Search in fieldsets.
    foreach ($form as &$element) {
      if (isset($element['#type']) && $element['#type'] == 'fieldset') {
        _drunkins_set_option_default($element, $element_name);
      }
    }
  }
}

/**
 * Submit callback.
 * Starts the batch update process.
 *
 * Note: various submit buttons may redirect here; some of them are defined
 * inside classes' settingsForm().
 */
function drunkins_batch_form_submit(&$form, &$form_state) {

  $job_id = $form_state['values']['drunkins_job_id'];
  // Usually we want to see messages on screen if we submit from a form. We
  // specify this unless the job config explicitly set it in 'runner_settings'.
  $info = drunkins_job_status_info($job_id);
  $logging = isset($info['runner_settings']['logging']) ? $info['runner_settings']['logging'] : DrunkinsJob::LOG_SCREEN | DrunkinsJob::LOG_WATCHDOG;

  // Get 'process settings' from all form values; clean out system values.
  $settings = array_diff_key($form_state['values'], array_flip(array(
    'form_build_id',
    'form_id',
    'form_token',
    'op',
    'submit',
    // ...plus values from submit buttons (which aren't settings)
    'cron',
    'display',
    // ...plus settings we know we don't need in $settings.
    'drunkins_job_id',
    'force_cron',
  )));
  // @todo remove 'logging' from settings, and have it only in runner_settings,
  //   when log() gets rewritten or removed from the job class. Also remove it
  //   from the job settingsForm() if possible; it is only there because we
  //   don't have another easy way to
  $settings['logging'] = $logging;
  // We assume most 'runner settings' are not in this form (unless hardcoded
  // below), so start off with an almost empty array. (Any settings defined in
  // queue info hooks will be merged in drunkins_start_job().)
  $runner_settings = array(
    'logging' => $logging
  );
  $reset_logger_settings = NULL;

  switch ($form_state['input']['op']) {
    case t('Reset job'):
      // Some things can never happen on a form submit: just throw.
      if (empty($form_state['values']['drunkins_job_id'])) {
        throw new LogicException("Missing 'drunkins_job_id' value.");
      }
      if (!user_access('access drunkins advanced')) {
        throw new LogicException("Permission missing.");
      }
      module_load_include('inc', 'drunkins', 'drunkins.run');
      _drunkins_finish_process($form_state['values']['drunkins_job_id'], array('target_state' => 'manual_reset'));
      break;

    case t('Un-claim items'):
      // Some things can never happen on a form submit: just throw.
      if (empty($form_state['values']['drunkins_job_id']) || empty($form_state['values']['unclaim_items_max'])) {
        throw new LogicException("Missing 'drunkins_job_id'/'unclaim_items_max' value.");
      }
      if (!user_access('access drunkins advanced')) {
        throw new LogicException("Permission missing.");
      }
      db_update('queue')->fields(['expire' => 0])->condition('expire', $form_state['values']['unclaim_items_max'], '<=')
        ->execute();
      break;

    case t('Enqueue'):
      if (!empty($form_state['values']['force_cron'])) {
        $runner_settings['force_cron'] = TRUE;
      }
      $reset_logger_settings = DrunkinsJob::LOG_WATCHDOG;
      // no break!
    case t('Start batch'):
      // We always want to see "There are no items to process." so override any
      // log severity for that message.
      $runner_settings['log_severity_no_items'] = isset($runner_settings['log_level']) ? $runner_settings['log_level'] : WATCHDOG_INFO;
      // 'list_format' could influence the output of start(), so clear too.
      unset ($settings['list_format']);
      $method = $form_state['input']['op'] == t('Start batch') ? 'batch' : 'queue';

      drunkins_start_job($job_id, $method, $runner_settings, $settings);

      if ($reset_logger_settings) {
        // For the remainder of the process, logging should be different. We
        // should be able to trust that our context is still 'current', not
        // saved to the database yet - so if we set the info, it should be
        // saved later. (This is a bit wishy-washy but it'll do until we
        // rewrite things.)
        $settings['logging'] = $reset_logger_settings;
        _drunkins_set_process_config($job_id, 'job', $settings);
        $runner_settings['logging'] = $reset_logger_settings;
        _drunkins_set_process_config($job_id, 'runner', $runner_settings);
      }
      break;

    default:
      module_load_include('inc', 'drunkins', 'drunkins.run');
      // We'd officially need to set logging in the 'runner settings' and
      // drunkins_get_job() would then move it to settings for us. But let's
      // 'cheat' and be more direct in this case by setting this unofficial
      // logger setting, since we don't need a runner. Maybe look at this later.
      $settings['logging'] = DrunkinsJob::LOG_SCREEN | DrunkinsJob::LOG_WATCHDOG;
      $job = drunkins_get_job($job_id, $settings);
      if (method_exists($job, 'settingsFormSubmit')) {
        $output = $job->settingsFormSubmit($form_state['input']['op']);

        // If the method returns a string, this is assumed to be HTML output.
        // We are however in the middle of a form submit function -- so we
        // will add this output to the form, and not redirect. The usual form
        // will be rendered below our output.
        if (is_string($output)) {
          $form['output'] = array(
            '#type'   => 'markup',
            '#markup' => '<h2>'
                         . t('The %a action produced the following output:',
                             array('%a' => $form_state['input']['op']))
                         . '</h2><hr/>' . $output . '<div class="description">&lt' . t('end output') . '&gt;</div><hr/>',
            '#weight' => -99,
          );
          $form_state['no_redirect'] = TRUE;
        }
      }
      else {
        drupal_set_message(t('Unknown operation %op!', array('%op' => $form_state['input']['op'])), 'error');
      }
  }
}


/**
 * Form builder for general Drunkins settings form.
 *
 * WARNING - This depends on the drunkins_process_run table existing, which
 * currently is not created automatically upon installing this module and
 * depends on the Data module to be able to create it (through a few clicks in
 * the admin UI). We'll fix that later - likely when we move the needed Views
 * integration into here. And that might be after I figured out and documented
 * the Joins stuff inside the data module, which is so ugly and/or buggy...
 *
 * (I may actually be completely mistaken and the Views integration in the
 * Data module may not be needed at all, for the exported view to work
 * correctly - in which case I can drop this dependency. For now, I'm too lazy
 * to spend time checking that.)
 */
function drunkins_general_settings_form($form, &$form_state) {
  // For the constant:
  module_load_include('inc', 'drunkins', 'drunkins.run');

  $form['drunkins_cron_time'] = array(
    '#type' => 'textfield',
    '#title' => t('Time to spend during cron runs'),
    '#description' => t('The available time will be divided among all running jobs - possibly excluding fetching items. Default is @default seconds.', ['@default' => DRUNKINS_DEFAULT_CRON_TIME]),
    '#default_value' => variable_get('drunkins_cron_time', NULL),
  );
  // We keep this option visible if Elysia isn't enabled, as a reminder to
  // install it and as indirect documentation why.
  $form['drunkins_cron_time_includes_start'] = array(
    '#type' => 'checkbox',
    '#title' => t('Starting/finishing jobs is included in cron time'),
    '#description' => t('If checked, the total time spent by drunkins in cron runs gets more predictable, by lowering the time available for actual processing of queued items when time is spent e.g. fetching items. This works <strong> only if Elysia cron is enabled</strong>.'),
    '#default_value' => variable_get('drunkins_cron_time_includes_start', FALSE),
  );
  $form['drunkins_save_process_summary'] = array(
    '#type' => 'checkbox',
    '#title' => t('Save process statistics'),
    '#description' => t('Save times & counters for every finished process, for statistics tracking. <strong>Warning:</strong> this depends on the Data module being enabled, and the standard view that comes with this module needs the Views Crosstab module with issues #3173306 and #3173309 patched, and possibly also Views Data Export.'),
    '#default_value' => variable_get('drunkins_save_process_summary', FALSE),
  );
  $form['drunkins_process_summary_checkpoint'] = array(
    '#type' => 'checkbox',
    '#title' => t('Save process checkpoint statistics'),
    '#description' => t('Save times & counters for every time cron picks up/pauses a process.'),
    '#default_value' => variable_get('drunkins_process_summary_checkpoint', FALSE),
    '#states' => [
      'enabled' => [
        ':input[name="drunkins_save_process_summary"]' => array('checked' => TRUE),
      ],
    ],
  );

  return system_settings_form($form);
}

/**
 * Form builder for Drinkins statistics deletion.
 */
function drunkins_delete_stats_form($form, &$form_state) {
  if (!db_table_exists('drunkins_process_run')) {
    return array('info' => array(
      '#type' => 'markup',
      '#markup' => t("The necessary statistics tables don't exist. For the moment, you need to install the Data module and create the tables through the admin interface (from definitions included in this module).")
    ));
  }

  $row = db_query("SELECT COUNT(*) as cnt, MIN(started) as min_started FROM {drunkins_process_run}")
    ->fetchAssoc();
  if (empty($row['cnt'])) {
    $form['info'] = array(
      '#type' => 'markup',
      '#markup' => t('The statistics tables are empty.'),
    );
  }
  else {
    $counter_count = db_query("SELECT COUNT(*) as cnt FROM {drunkins_process_counter}")
        ->fetchField();
    $form['info'] = array(
      '#type' => 'markup',
      '#markup' => t('The statistics tables contain @count1 process/checkpoint values with @count2 counter values, starting @date.',
      array('@count1' => $row['cnt'], '@date' => format_date($row['min_started']), '@count2' => $counter_count)
      ),
    );
  }
  $form['delete_before'] = array(
    '#type' => 'textfield',
    '#title' => t('Delete statistics before'),
    '#required' => TRUE,
    '#description' => t('Format: anything PHP can parse.')
  );
  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Delete'),
  );

  return $form;
}

function drunkins_delete_stats_form_validate(&$form, &$form_state) {
  $timestamp = strtotime($form_state['values']['delete_before']);
  if (!$timestamp) {
    form_set_error('delete_before', t('Cannot parse date.'));
  }
}

function drunkins_delete_stats_form_submit(&$form, &$form_state) {
  $timestamp = strtotime($form_state['values']['delete_before']);
  db_delete('drunkins_process_run')
    ->condition('started', $timestamp, '<')
    ->execute();
  // Now-orphaned counters. (If we were forcing foreign keys at the database
  // level, we'd need to swap the order.)
  db_query('DELETE del
    FROM {drunkins_process_counter} del LEFT JOIN drunkins_process_run r ON del.process_id=r.id
    WHERE r.id IS NULL')->execute();
}
